C
C     ELLBDY BUILD BOUNDARY INTEGRALS 
C*************************************************************************
C
      SUBROUTINE ELLBDY(NSCOL,NSROW,NSLAY,NCOL,NROW,NLAY,
     *                 CINFL,RHS,RHSO,NODESS,IABOVE,IBELOW,NCINFL,
     *                 IDBNDY,CTCFLO,LOCBDY,LBNDY,DIAGO,NFACES,NTFACE,
     *                 VC,VR,VL,RF,
     *                 THCK,POR,IBOUND,
     *                 IOUTS,TIMV,TSTEP2,VCMAX,VRMAX,VLMAX,
     *                 DELCOL,DELROW,XFOR,XBAC,YFOR,YBAC,DECAY,SRCDCY)
C
C*************************************************************************
C
C  1.  BUILD LBNDY TABLE.  FOR EACH BOUNDARY FACE,
C      LBNDY(FACE NBR)= -1       IF INFLOW            (CTCFLO>0)
C                     =  0       IF OUTFLOW           (CTCFLO<0)
C                     =CELL CENTER NODE NBR    IF NOFLOW 
C                                             (CTCFLO=0, OR NO CTCFLO ENTRY)
C                     = -2       IF FACE OF INACTIVE CELL
C
C  2.  FOR EACH OUTFLOW BOUNDARY FACE, FIND VOL ACROSS FACE DURING TIME STEP:
C       DIAGO(FACE NBR)=-CTCFLO(ID)*TIMV
C
C  3.  FOR EACH INFLOW BOUNDARY FACE,
C      SPACIALLY DISCRETIZE FACE
C      FORWARD TRACK INTEGRATION POINT TO NEW TIME LEVEL
C      ADD INFLOW MASS TO RHS OF EQN FOR DESTINATION CELL OF FORW TRACKED POINT:
C      RHS(DEST NODE)=CINFL*CTCFLO/(POR*RF)*FRAC IN CELL*EXP(-DT*DECAY),
C                     INTEGRATED USING REFERENCE AREA AND TIME
C                     USING TRAP INTEGRATION RULE
C                    =MASS IN/POROSITY
C      RHSO          =MASS IN
C                     (DON'T DIVIDE OUT POROSITY)
C
C  4.  BUILD IDBNDY TABLE.  FOR EACH CTCFLO ENTRY,
C      IDBNDY(CTCFLO ID)= -FACE NBR       IF INFLOW
C                       =  FACE NBR       IF OUTFLOW
C                       =  0              IF NOFLOW
C
C*************************************************************************
C
C  BUILD LBNDY TABLE.  ENTRY FOR EACH BOUNDARY FACE.  FACE NUMBERING IS 
C  IDENTICAL TO CTCFLO ID NUMBERING FOR CASE WHERE TRANSPORT DOMAIN IS
C  STRICTLY WITHIN FLOW DOMAIN ON ALL SIDES: NUMBERING IS
C   XFACES, YFACES, ZFACES,
C   X IS LEFT, RIGHT ACROSS ROW THEN LAYER
C   Y IS BACK, FRONT ACROSS COL THEN LAYER
C   Z IS TOP, BOTTOM ACROSS COL THEN ROW
C
C  ID=INDEX INTO CTCFLO TABLE CONTAINING FLOW RATES FOR FACES
C        -  NOT ON BOUNDARY OF FLOW GRID
C        -  NOT IN OMITTED DIMENSION OF 1-DIM OR 2-DIM PROBLEM
C
C  LBNDY IS CONSTRUCTED FOR ALL CELL FACES, ONE DIRECTION AT A TIME
C
C  IDBNDY IS CONSTRUCTED FOR ALL CTCFLO ENTRIES, AS LBNDY IS BUILT
C
C  ISTRT=STARTING INDEX INTO LBNDY ARRAY FOR DIRECTION
C  IEND=ENDING INDEX INTO LBNDY ARRAY FOR DIRECTION
C  IPTR1,2=CELL CENTER NODE NUMBER FOR BOUNDARY CELL
C  IDEL=CELL CENTER NODE NBR INCREMENT BETWEEN BOUNDARY FACES
C  INCRE=NBR OF CELL FACES THIS DIRECTION 
C  NBRANCH=BRANCH ADDRESS POINTER FOR PROCESSING A DIRECTION 
C          FOR INFLOW MASS
C  LBRANCH=BRANCH ADDRESS FOR RETURN AFTER IN,OUT,NOFLOW EXECUTION
C************************************************************************
C
      DOUBLE PRECISION DECAY
      COMMON /SUBGRD/ ISCOL1,ISCOL2,ISROW1,ISROW2,ISLAY1,
     *  ISLAY2,ISUBGD
      COMMON /ELLAM/ CINV,RINV,BINV,HCINV,HRINV,HBINV,
     *               STINIT,ADINIT,STMASS,ADMASS,OLMASS,
     *               AZERO,
     *               NSC,NSR,NSL,NT,NCTF,NRTF,NLTF,
     *               NEIGHB(8,2),NSLOPE(3),
     *               IDTOP,IDMAX,NCOEF,NSCH,NSRH,NSLH
      dimension rhs(nodess),ctcflo(nfaces),cinfl(ncinfl)
C
      NSCR=NSCOL*NSROW
C
C  BUILD TABLES FOR EACH PARALLEL PAIR OF FACES
C
C  X FACES:
      ID=0
      ISTRT=1
      IEND=NSROW*NSLAY*2
      NBRANCH=1
      IDEL=NSCOL
      IPTR1=1
      IPTR2=NSCOL
      LIM1=ISCOL1
      LIM2=ISCOL2
      MAX=NCOL
      CALL BDYENT(ISTRT,IEND,LIM1,LIM2,ID,IPTR1,IPTR2,IDEL,
     *           MAX,NBRANCH,CTCFLO,LOCBDY,
     *           IDBNDY,NFACES,LBNDY,NTFACE,IABOVE,IBELOW,
     *           DIAGO,CINFL,NCINFL,CONCPR,NSCOL,NSROW,NSLAY,
     *           VC,VR,VL,RF,THCK,POR,IBOUND,NCOL,NROW,NLAY,
     *           IOUTS,TIMV,VCMAX,VRMAX,VLMAX,NODESS,RHS,RHSO,
     *           DELCOL,DELROW,XFOR,XBAC,YFOR,YBAC,DECAY,SRCDCY)
C
C  Y FACES
      DO 20 M=1,NSLAY
        ISTRT=IEND+1
        IEND=IEND+NSCOL*2
        NBRANCH=2
        IDEL=1
        IPTR1=1+NSCR*(M-1)
        IPTR2=NSCR*M-NSCOL+1
        LIM1=ISROW1
        LIM2=ISROW2
        MAX=NROW
        CALL BDYENT(ISTRT,IEND,LIM1,LIM2,ID,IPTR1,IPTR2,IDEL,
     *           MAX,NBRANCH,CTCFLO,LOCBDY,
     *           IDBNDY,NFACES,LBNDY,NTFACE,IABOVE,IBELOW,
     *           DIAGO,CINFL,NCINFL,CONCPR,NSCOL,NSROW,NSLAY,
     *           VC,VR,VL,RF,THCK,POR,IBOUND,NCOL,NROW,NLAY,
     *           IOUTS,TIMV,VCMAX,VRMAX,VLMAX,NODESS,RHS,RHSO,
     *           DELCOL,DELROW,XFOR,XBAC,YFOR,YBAC,DECAY,SRCDCY)
C
20    CONTINUE
C
C  Z FACES
      ISTRT=IEND+1
      IEND=IEND+NSCR*2
      NBRANCH=3
      IDEL=1
      IPTR1=1
      IPTR2=NSCR*(NSLAY-1)+1
      LIM1=ISLAY1
      LIM2=ISLAY2
      MAX=NLAY
      CALL BDYENT(ISTRT,IEND,LIM1,LIM2,ID,IPTR1,IPTR2,IDEL,
     *           MAX,NBRANCH,CTCFLO,LOCBDY,
     *           IDBNDY,NFACES,LBNDY,NTFACE,IABOVE,IBELOW,
     *           DIAGO,CINFL,NCINFL,CONCPR,NSCOL,NSROW,NSLAY,
     *           VC,VR,VL,RF,THCK,POR,IBOUND,NCOL,NROW,NLAY,
     *           IOUTS,TIMV,VCMAX,VRMAX,VLMAX,NODESS,RHS,RHSO,
     *           DELCOL,DELROW,XFOR,XBAC,YFOR,YBAC,DECAY,SRCDCY)
C
C
      RETURN
      END
C
C********************************************************************************
C
      SUBROUTINE BDYENT(ISTRT,IEND,LIM1,LIM2,ID,IPTR1,IPTR2,IDEL,
     *           MAX,NBRANCH,CTCFLO,LOCBDY,
     *           IDBNDY,NFACES,LBNDY,NTFACE,IABOVE,IBELOW,
     *           DIAGO,CINFL,NCINFL,CONCPR,NSCOL,NSROW,NSLAY,
     *           VC,VR,VL,RF,THCK,POR,IBOUND,NCOL,NROW,NLAY,
     *           IOUTS,TIMV,VCMAX,VRMAX,VLMAX,NODESS,RHS,RHSO,
     *           DELCOL,DELROW,XFOR,XBAC,YFOR,YBAC,DECAY,SRCDCY)
C
C  UPDATES BOUNDARY ARRAYS LBNDY AND IDBNDY
C  CALCULATES DIAGO ENTRY FOR OUTFLOW FACES 
C  ENTRY IS WITH POINTERS SET
C  FOR X, Y, OR Z FACES
C********************************************************************************
      DOUBLE PRECISION DECAY
      DIMENSION CTCFLO(NFACES),LBNDY(NTFACE),IDBNDY(NFACES),
     *         LOCBDY(3,NFACES),CINFL(NCINFL),DIAGO(NTFACE),
     *         IBOUND(NCOL,NROW,NLAY),RF(NSCOL,NSROW,NSLAY)
      COMMON /SUBGRD/ ISCOL1,ISCOL2,ISROW1,ISROW2,ISLAY1,
     *  ISLAY2,ISUBGD
      COMMON /ELLAM/ CINV,RINV,BINV,HCINV,HRINV,HBINV,
     *               STINIT,ADINIT,STMASS,ADMASS,OLMASS,
     *               AZERO,
     *               NSC,NSR,NSL,NT,NCTF,NRTF,NLTF,
     *               NEIGHB(8,2),NSLOPE(3),
     *               IDTOP,IDMAX,NCOEF,NSCH,NSRH,NSLH
C
      DO 100 N1=ISTRT,IEND,2
        ASSIGN 15 TO LBRANCH
        N=N1
        IF (LIM1.GT.1) THEN
C
C  CTFLO ENTRY EXISTS, PROCESS OUT-, NO- OR INFLOW
          ID=ID+1
          IF (CTCFLO(ID)) 300,400,500
        ELSE 
C
C  FACES FOR WHICH NO CTCFLO ENTRY EXIST
C  (TRANSPORT SUBGRID ALIGNED WITH FLOW GRID BOUNDARY)
C  ARE NOFLOW FACES
          LBNDY(N)=IPTR1
C
C  IF CELL IS INACTIVE, INDICATE IN LBNDY
          JS=MOD(IPTR1-1,NSCOL)+1
          J=ISCOL1+JS-1
          KS=(IPTR1-1)/(NSCOL*NSROW)+1
          K=ISLAY1+KS-1
          IS=(IPTR1-JS)/NSCOL-(KS-1)*NSROW+1
          I=ISROW1+IS-1
          IF (IBOUND(J,I,K).EQ.0) LBNDY(N)=-2
        ENDIF
C
15    IPTR1=IPTR1+IDEL
      ASSIGN 25 TO LBRANCH
      N=N1+1 
C
      IF (LIM2.LT.MAX) THEN
        ID=ID+1
        IF (CTCFLO(ID)) 300,400,500
      ELSE
        LBNDY(N)=IPTR2
        JS=MOD(IPTR2-1,NSCOL)+1
        J=ISCOL1+JS-1
        KS=(IPTR2-1)/(NSCOL*NSROW)+1
        K=ISLAY1+KS-1
        IS=(IPTR2-JS)/NSCOL-(KS-1)*NSROW+1
        I=ISROW1+IS-1
        IF (IBOUND(J,I,K).EQ.0) LBNDY(N)=-2
      ENDIF
C
25      IPTR2=IPTR2+IDEL
      GOTO 100
C
C  FACES WITH CTCFLO ENTRY
C
C  OUTFLOW
C
300   LBNDY(N)=0
      IDBNDY(ID)=N
      DIAGO(N)=-CTCFLO(ID)*TIMV
      GOTO 200
C
C  NO FLOW
C
400   LBNDY(N)=(LOCBDY(3,ID)-ISLAY1)*NSROW*NSCOL
     *           +(LOCBDY(2,ID)-ISROW1)*NSCOL
     *           +LOCBDY(1,ID)-ISCOL1+1
C
C  IF CELL IS INACTIVE, INDICATE IN LBNDY
          JS=MOD(LBNDY(N)-1,NSCOL)+1
          J=ISCOL1+JS-1
          KS=(LBNDY(N)-1)/(NSCOL*NSROW)+1
          K=ISLAY1+KS-1
          IS=(LBNDY(N)-JS)/NSCOL-(KS-1)*NSROW+1
          I=ISROW1+IS-1
          IF (IBOUND(J,I,K).EQ.0) LBNDY(N)=-2
C
      IDBNDY(ID)=0
      GOTO 200
C
C  INFLOW
C
500   J=LOCBDY(1,ID)
      I=LOCBDY(2,ID)
      K=LOCBDY(3,ID)
C      
C   DISOLVED SOLUTE CONCENTRATION ON BNDY IS CINFL/RF
C   IDTOP IS PLACED IN ELLAM.INC BY SMOC5BY
C
      JS=J-ISCOL1+1
      IS=I-ISROW1+1
      KS=K-ISLAY1+1
      IF (ID.LT.IDTOP) THEN
         CPRIME=CINFL(KS)/RF(JS,IS,KS)
      ELSE
            IF (NSROW.EQ.NROW.AND.NSCOL.EQ.NCOL) THEN       ! Added by YCLin
               NNSLAY = 0                                   !       :
            ELSE IF (NSROW.NE.NROW.OR.NSCOL.NE.NCOL) THEN   !       :
               NNSLAY = NSLAY                               !       :
            END IF                                          !       :
          CPRIME=0.0
         IF (MOD(ID,2).EQ.0) THEN
            IF (IABOVE.EQ.0.AND.IBELOW.EQ.1)
     *                      CPRIME=CINFL(NNSLAY+1)/RF(JS,IS,KS)
            IF (IABOVE.EQ.1.AND.IBELOW.EQ.1)
     *                      CPRIME=CINFL(NNSLAY+2)/RF(JS,IS,KS)
         ELSE
              IF (IABOVE.EQ.1) CPRIME=CINFL(NNSLAY+1)/RF(JS,IS,KS)
         END IF
      END IF
C
      CONCPR=CPRIME*CTCFLO(ID)
      LBNDY(N)=-1
      IDBNDY(ID)=-N
C
      IF (CONCPR.NE.0.0) GOTO(510,520,530),NBRANCH
      GOTO 200
C
C X FACE
510   AREA=RINV*BINV
      XHAT=-.5
      ZHAT=-.5-HBINV
      DO 700 NXL=1,NSL
        ZHAT=ZHAT+BINV
C
C
        YHAT=-.5-HRINV
        DO 700 NXR=1,NSR
          YHAT =YHAT+RINV
C
C
          CALL BDYRHS(VC,VR,VL,RF,
     *           THCK,POR,IBOUND,
     *           NCOL,NROW,NLAY,NSCOL,NSROW,NSLAY,
     *           IOUTS,TIMV,VCMAX,VRMAX,VLMAX,
     *           NODESS,LBNDY,NTFACE,CONCPR,AREA,
     *           J,I,K,XHAT,YHAT,ZHAT,RHS,RHSO,
     *           DELCOL,DELROW,XFOR,XBAC,YFOR,YBAC,DECAY,SRCDCY)
C
700    CONTINUE
      GOTO 200
C
C Y FACE
520   AREA=CINV*BINV
      YHAT=-.5
      ZHAT=-.5-HBINV
      DO 800 NYL=1,NSL
        ZHAT=ZHAT+BINV
C
C
        XHAT=-.5-HCINV
        DO 800 NYC=1,NSC
          XHAT=XHAT+CINV
C
          CALL BDYRHS(VC,VR,VL,RF,
     *           THCK,POR,IBOUND,
     *           NCOL,NROW,NLAY,NSCOL,NSROW,NSLAY,
     *           IOUTS,TIMV,VCMAX,VRMAX,VLMAX,
     *           NODESS,LBNDY,NTFACE,CONCPR,AREA,
     *           J,I,K,XHAT,YHAT,ZHAT,RHS,RHSO,
     *           DELCOL,DELROW,XFOR,XBAC,YFOR,YBAC,DECAY,SRCDCY)
C
800   CONTINUE
      GOTO 200
C
C Z FACE
530   AREA=CINV*RINV
      ZHAT=-.5
      YHAT=-.5-HRINV
      DO 900 NZR=1,NSR
        YHAT=YHAT+RINV
C
C
        XHAT=-.5-HCINV
        DO 900 NZC=1,NSC
          XHAT=XHAT+CINV
C
C
          CALL BDYRHS(VC,VR,VL,RF,
     *           THCK,POR,IBOUND,
     *           NCOL,NROW,NLAY,NSCOL,NSROW,NSLAY,
     *           IOUTS,TIMV,VCMAX,VRMAX,VLMAX,
     *           NODESS,LBNDY,NTFACE,CONCPR,AREA,
     *           J,I,K,XHAT,YHAT,ZHAT,RHS,RHSO,
     *           DELCOL,DELROW,XFOR,XBAC,YFOR,YBAC,DECAY,SRCDCY) 
900   CONTINUE
200   GOTO LBRANCH
C
100   CONTINUE
      RETURN
      END
C
C**************************************************************************
C
      SUBROUTINE BDYRHS(VC,VR,VL,RF,
     *           THCK,POR,IBOUND,
     *           NCOL,NROW,NLAY,NSCOL,NSROW,NSLAY,
     *           IOUTS,TIMV,VCMAX,VRMAX,VLMAX,
     *           NODESS,LBNDY,NTFACE,CONCPR,AREA,
     *           J,I,K,XHAT,YHAT,ZHAT,RHS,RHSO,
     *           DELCOL,DELROW,XFOR,XBAC,YFOR,YBAC,DECAY,SRCDCY)
C**************************************************************************
C      FOR EACH INFLOW BOUNDARY FACE,
C      SPACIALLY DISCRETIZE FACE
C      FORWARD TRACK INTEGRATION POINT TO NEW TIME LEVEL
C      ADD INFLOW MASS TO RHS OF EQN FOR DESTINATION CELL OF FORW TRACKED POINT:
C      RHS(DEST NODE)=CINFL*CTCFLO/(POR*RF)*FRAC IN CELL*EXP(-DT*DECAY),
C                     INTEGRATED USING REFERENCE AREA AND TIME
C                     USING TRAP INTEGRATION RULE
C                    =MASS IN/POROSITY
C      RHSO          =MASS IN
C                     (DON'T DIVIDE OUT POROSITY)
C**************************************************************************      
C
      DOUBLE PRECISION FACT,DECAY
      DIMENSION TFVAL(8),NUMTF(8),NSCOORD(3,8)
      DIMENSION RHS(NODESS),RHSO(NTFACE),POR(NODESS),RF(NODESS)
      COMMON /SUBGRD/ ISCOL1,ISCOL2,ISROW1,ISROW2,ISLAY1,
     *  ISLAY2,ISUBGD
      COMMON /ELLAM/ CINV,RINV,BINV,HCINV,HRINV,HBINV,
     *               STINIT,ADINIT,STMASS,ADMASS,OLMASS,
     *               AZERO,
     *               NSC,NSR,NSL,NT,NCTF,NRTF,NLTF,
     *               NEIGHB(8,2),NSLOPE(3),
     *               IDTOP,IDMAX,NCOEF,NSCH,NSRH,NSLH
      dimension thck(nodess)
C
      NSCR=NSCOL*NSROW
      PROD=CONCPR*AREA
      PC=J+XHAT
      PR=I+YHAT
      PL=K+ZHAT
C
C  FIND PARTIAL TIME INCREMENT, OVER WHICH TO TRACK, DECAY MASS 
      TINC=TIMV/NT
C
      DT=TIMV-tinc/2
      DO 100 N=1,NT
C
      PCMV=PC
      PRMV=PR
      PLMV=PL
      CALL ELMOVE(PCMV,PRMV,PLMV,VC,VR,VL,RF,
     *           THCK,POR,IBOUND,
     *           NCOL,NROW,NLAY,NSCOL,NSROW,NSLAY,
     *           IOUTS,DT,TSTEP2,VCMAX,VRMAX,VLMAX,
     *           JSMV,ISMV,KSMV,JMV,IMV,KMV,NLOC,DELCOL,DELROW)
C
C   EVALUATE ALL NONZERO TEST FUNCTIONS AT FORWARD TRACKED POINT
C   AT NEW TIME LEVEL
C
c      noda=(ksmv-1)*nscol*nsrow+(ismv-1)*nscol+jsmv
      IF (NLOC.EQ.0) THEN
                CALL EVALTF(PCMV,PRMV,PLMV,TFVAL,IBOUND,
     *           NUMTF,JSMV,ISMV,KSMV,JMV,IMV,KMV,NCOL,NROW,NLAY,
     *           NSCOL,NSROW,NSLAY,NNZER,NSCOORD,THCK,XFOR,XBAC,
     *           YFOR,YBAC)
      ELSE
                CALL BNDYTF(PCMV,PRMV,PLMV,TFVAL,IBOUND,
     *           NUMTF,JSMV,ISMV,KSMV,JMV,IMV,KMV,NCOL,NROW,NLAY,
     *           NSCOL,NSROW,NSLAY,NODESS,NNZER,LBNDY,NTFACE,THCK,
     *           XFOR,XBAC,YFOR,YBAC)
      ENDIF
C
c      nnzer=1
c      tfval(1)=1d0
c      numtf(1)=noda

C
      TPROD=TINC*PROD


c      IF (N.EQ.1) TPROD=.5*TINC*PROD


C
C   IF DECAY IS NONZERO, THE FOLLOWING WILL YIELD DECAY FACTOR(<>1)
C   AND ACCUMULATE DECAYED MASS FOR MASS BALANCE
C
      FACT=DEXP(-DT*DECAY)
      SRCDCY=SRCDCY+(FACT*TPROD-TPROD)
C      
      DO 600 M=1,NNZER
C
      IF (NUMTF(M).LE.NODESS) THEN
         PRINV=1/POR(NUMTF(M))
         RHS(NUMTF(M))=RHS(NUMTF(M))+FACT*TFVAL(M)*TPROD*PRINV
cgzh       if (NUMTF(M).eq.18) print *,'ellbdy',' rhs',
cgzh     *       RHS(NUMTF(M)),' added',FACT*TFVAL(M)*TPROD*PRINV
      ELSE
         RHSO(NUMTF(M)-NODESS)=RHSO(NUMTF(M)-NODESS)
     *                           +FACT*TFVAL(M)*TPROD
C
      ENDIF
600   CONTINUE
C
      DT=DT-TINC
100   CONTINUE
C
C  NOW DO SPACIAL POINTS WHICH ENTER SUBGRID AT TIME=NEW TIME,
C  AND THUS DON'T MOVE DURING THIS TIME STEP
c      JS=J-ISCOL1+1
c      IS=I-ISROW1+1
c      KS=K-ISLAY1+1
c      CALL EVALTF(PC,PR,PL,TFVAL,IBOUND,
c     *           NUMTF,JS,IS,KS,J,I,K,NCOL,NROW,NLAY,
c     *           NSCOL,NSROW,NSLAY,NNZER,NSCOORD,THCK,
c     *           XFOR,XBAC,YFOR,YBAC)
c      TPROD=.5*TINC*PROD
c      DO 601 M=1,NNZER
c      PRINV=1/POR(NUMTF(M))
c      RHS(NUMTF(M))=RHS(NUMTF(M))+TFVAL(M)*TPROD*PRINV
c      if (NUMTF(M).eq.18) print *,'ellbdy inflo at 18'


601   CONTINUE
      RETURN
      END
